import { IdToken } from './scan.js';
import { ValueType } from './value.js';
import WasmNumber from './numbers.js';
import * as error from './error.js';
/**
 * Abstract base for datatypes
 */
export class Type {
    token;
    /**
     * @param {LexerToken} [token] -
     */
    constructor(token = undefined) {
        this.token = token;
    }
}
/**
 * Type which matches any other type
 */
export class AnyType extends Type {
    /**
     * @override
     */
    check() {
        return true;
    }
    flatPrimitiveList() {
        throw new error.SyntaxError('AnyType can only exist at compile-time', [this.token]);
    }
    getWasmTypeName() {
        throw new error.SyntaxError('AnyType can only exist at compile-time', [this.token]);
    }
    isUnit() {
        throw new error.SyntaxError('AnyType can only exist at compile-time', [this.token]);
    }
    toString() {
        return "Any";
    }
}
/**
 * Type variables - idk might not even implement this
 * ie: (($A $A $B): ( $v1 $v2 $obj ) = "field" v1 v2 + obj JSON.withMember )
 */
export class TypeVarType extends Type {
    identifier;
    scope;
    type;
    constructor(token, identifier, scope) {
        super(token);
        this.identifier = identifier;
        this.scope = scope;
    }
    check(type) {
        if (type instanceof ClassType)
            type = type.getBaseType();
        if (!type)
            return false;
        if (type instanceof AnyType)
            return true;
        if (type instanceof TypeVarType)
            return this.identifier === type.identifier && this.scope === type.scope;
        if (this.type)
            return this.type.check(type);
        return true;
    }
    toString() {
        return '$' + this.identifier;
    }
    flatPrimitiveList() {
        throw new error.SyntaxError('TypeVarType can only exist at compile-time', [this.token]);
    }
    isUnit() {
        throw new error.SyntaxError('TypeVarType can only exist at compile-time', [this.token]);
    }
    getWasmTypeName(name) {
        throw new error.SyntaxError('TypeVarType can only exist at compile-time', [this.token]);
    }
}
/**
 * Type which never matches
 */
export class NeverType extends Type {
    /**
     * @override
     */
    check() {
        return false;
    }
    flatPrimitiveList() {
        throw new error.SyntaxError('NeverType can only exist at compile-time', [this.token]);
    }
    getWasmTypeName() {
        throw new error.SyntaxError('NeverType can only exist at compile-time', [this.token]);
    }
    isUnit() {
        throw new error.SyntaxError('NeverType can only exist at compile-time', [this.token]);
    }
    toString() {
        return "Never";
    }
}
/**
 * When need to be able to handle more than one type
 *  aka sum type (`|`)
 */
// TODO when tracing this need to trace all possible types
export class UnionType extends Type {
    types;
    constructor(token = undefined, types = []) {
        super(token);
        this.types = types;
    }
    /**
     * @override
     */
    check(type) {
        if (type === null)
            return false;
        // Don't care about classes
        if (type instanceof ClassType) {
            type = type.getBaseType();
            if (!type)
                return false;
        }
        // Verify other union is subset of this one
        if (type instanceof UnionType)
            return type.types.every(t => this.types.includes(t));
        if (type instanceof AnyType)
            return true;
        if (type instanceof NeverType)
            return false;
        // Verify type is in this set
        return this.types.some(t => t.check(type));
    }
    flatPrimitiveList() {
        throw new error.SyntaxError('UnionType can only exist at compile-time', [this.token]);
    }
    getWasmTypeName() {
        throw new error.SyntaxError('UnionType can only exist at compile-time', [this.token]);
    }
    isUnit() {
        throw new error.SyntaxError('UnionType can only exist at compile-time', [this.token]);
    }
    toString() {
        let ret = this.types[0].toString();
        for (let i = 1; i < this.types.length; i++)
            ret += ` ${this.types[i].toString()} |`;
        return ret;
    }
}
/**
 * Syntactic types
 */
export class SyntaxType extends Type {
    valueType;
    /**
     * @param token location in code
     * @param valueType relevant syntactic/value type
     */
    constructor(token, valueType) {
        super(token);
        this.valueType = valueType;
    }
    /**
     * Singletons for each valuetype
     */
    static ValueTypes = {
        [ValueType.Macro]: new SyntaxType(new IdToken('Syntax:Macro', 0, undefined), ValueType.Macro),
        [ValueType.Data]: new SyntaxType(new IdToken('Syntax:Data', 0, undefined), ValueType.Data),
        [ValueType.Type]: new SyntaxType(new IdToken('Syntax:Type', 0, undefined), ValueType.Type),
        [ValueType.Id]: new SyntaxType(new IdToken('Syntax:Id', 0, undefined), ValueType.Id),
        [ValueType.Expr]: new SyntaxType(new IdToken('Syntax:Expr', 0, undefined), ValueType.Expr),
        [ValueType.Fxn]: new SyntaxType(new IdToken('Syntax:Fxn', 0, undefined), ValueType.Fxn),
        [ValueType.Str]: new SyntaxType(new IdToken('Syntax:Str', 0, undefined), ValueType.Str),
        [ValueType.Ns]: new SyntaxType(new IdToken('Syntax:Ns', 0, undefined), ValueType.Ns),
    };
    /**
     * @override
     */
    check(type) {
        // Drop classes
        if (type instanceof ClassType)
            type = type.getBaseType();
        // Match any
        if (type instanceof AnyType)
            return true;
        // Compare singletons
        if (type instanceof SyntaxType)
            return type.valueType === this.valueType;
        // Note: don't allow classes of syntax types?
        return false;
    }
    /**
     * @overrride
     */
    toString() {
        return 'Syntax:' + ValueType[this.valueType];
    }
}
// This is bad because tuple types will have type Syntax:Type
// export class AnyTupleType extends Type {
//     protected constructor() {
//         super(new IdToken('Tuple', 0, undefined));
//     }
//     static readonly singleton = new AnyTupleType();
//     check(type: Type): boolean {
//         // Drop classes
//         if (type instanceof ClassType)
//             type = type.getBaseType();
//         // Typecheck
//         return type instanceof TupleType
//             || type instanceof AnyTupleType
//             || type instanceof AnyType;
//     }
//     toString(): string {
//         return 'Tuple';
//     }
// }
/**
 * Types of entities that can be physically represented
 */
export class DataType extends SyntaxType {
    recursive;
    // declare valueType: ValueType.Data;
    constructor(token, recursive = false) {
        super(token, ValueType.Data);
        this.recursive = recursive;
    }
    /**
     * Gives the wasm typename for given type
     * @returns - typename
     * @virtual
     */
    getWasmTypeName(name) {
        return this.flatPrimitiveList().map(t => t instanceof PrimitiveType ? t.name : 'i32').join(' ');
    }
    /**
     * Can input/output values of this type be ignored by compiler (no value held)
     * @returns true if the value doesn't carry a value
     * @virtual
     */
    isUnit() {
        return this.flatPrimitiveList().length === 0;
    }
    /**
     * Number of i32's needed to store the described datastructure
     * @returns size in words
     */
    size() {
        return this.flatPrimitiveList().reduce((a, t) => a += t.size(), 0);
    }
}
/**
 * More specific than types, used for applying methods and stuff
 */
export class ClassType extends DataType {
    /**
     * Unique identifier for each class
     */
    id;
    static _tokenIdMap = new Map();
    /**
     * @param token - Code where
     * @param type - Underlying Data type
     * @param [id] - Clone a class
     * @param recursive - Does this class need to be stored on the heap?
     */
    constructor(token, type, id, recursive = false) {
        super(token, recursive);
        this.type = type;
        this.id = id === undefined ? genId() : id;
        function genId() {
            // Already in the map
            const id = ClassType._tokenIdMap.get(token);
            if (id !== undefined)
                return id;
            // Add to map
            const newId = ClassType._tokenIdMap.size;
            ClassType._tokenIdMap.set(token, newId);
            return newId;
        }
    }
    /**
     * @override
     */
    check(type) {
        if (type === null)
            return false;
        if (type instanceof AnyType)
            return true;
        // Check class
        if (type instanceof ClassType) {
            // TODO alternatively we can check if tokens align and then verify data schema
            // Check if subclass
            if (!type.getClassIds().includes(this.id))
                return false;
            // Check data schema
            if (!this.type)
                return true;
            return this.getBaseType().check(type);
        }
        // Expected a class
        return false;
    }
    /**
     * Gives type that class refers to
     * @returns - Underlying datatype
     */
    getBaseType() {
        let ret = this.type;
        while (ret instanceof ClassType)
            ret = ret.type;
        return ret;
    }
    /**
     * @override
     */
    flatPrimitiveList() {
        return this.getBaseType().flatPrimitiveList();
    }
    /**
     * @override
     */
    getWasmTypeName(name) {
        return this.getBaseType().getWasmTypeName(name);
    }
    /**
     * @returns - list of class ids for this type
     */
    getClassIds() {
        const ret = [this.id];
        let type = this.type;
        while (type instanceof ClassType) {
            ret.push(type.id);
            type = type.type;
        }
        return ret;
    }
    /**
     * @override
     */
    isUnit() {
        return this.getBaseType().isUnit();
    }
    toString() {
        return `${this.type.toString()} ${this.id} #class`;
    }
    /**
     * Does an object of this type need to be stored on the heap?
     */
    isRecursive() {
        return this.recursive
            || (this.type instanceof ClassType && this.type.isRecursive())
            || (this.type instanceof EnumBaseType && this.type.isRecursive())
            || (this.type instanceof EnumClassType && this.type.parent.isRecursive());
    }
}
/**
 * When need to store more than one piece of data in a single value
 *  aka product type (`pack`)
 */
export class TupleType extends DataType {
    types;
    /**
     * @param token location in code
     * @param types member types
     */
    constructor(token = undefined, types = []) {
        super(token);
        this.types = types;
    }
    /**
     * Returns an error if type contains compile-time-only member types
     */
    assertIsDataType() {
        const t = this.types.find(t => !(t instanceof DataType));
        if (t)
            return new error.SyntaxError('Unexpected compile-only type', [t.token, this.token]);
    }
    // /**
    //  * @override
    //  */
    // getWasmTypeName(name?: string) {
    //     const err = this.assertIsDataType();
    //     if (err)
    //         throw err;
    //     return (this.types as DataType[]).map(t => t.getWasmTypeName(name)).join(' ');
    // }
    /**
     * @override
     */
    isUnit() {
        return this.types.length === 0;
    }
    /**
     * @override
     */
    flatPrimitiveList() {
        const err = this.assertIsDataType();
        if (err)
            throw err;
        return this.types
            .map(t => t.flatPrimitiveList())
            .reduce((a, b) => a.concat(b), []);
    }
    /**
     * @override
     */
    check(type) {
        if (type == null)
            return false;
        // Don't care about classes
        if (type instanceof ClassType) {
            type = type.getBaseType();
            if (!type)
                return false;
        }
        // Always match Any
        if (type instanceof AnyType)
            return true;
        // Not a tuple
        if (!(type instanceof TupleType))
            return false;
        // Different size
        if (this.types.length !== type.types.length)
            return false;
        // Verify types match
        for (let i = 0; i < this.types.length; i++)
            if (!this.types[i].check(type.types[i]))
                return false;
        return true;
    }
    getBaseType() {
        return this;
    }
    toString() {
        if (this.types.length === 0)
            return 'Unit';
        return `( ${this.types.map(t => t.toString()).join(' ')} )`;
    }
}
/**
 * Type that's a component of compilation target
 */
export class PrimitiveType extends DataType {
    name;
    _size;
    // Map of WASM primitive types
    static Types = {
        I32: new PrimitiveType('i32', 1),
        I64: new PrimitiveType('i64', 2),
        F32: new PrimitiveType('f32', 1),
        F64: new PrimitiveType('f64', 2),
    };
    static typeMap = {
        [WasmNumber.Type.I32]: PrimitiveType.Types.I32,
        [WasmNumber.Type.I64]: PrimitiveType.Types.I64,
        [WasmNumber.Type.F32]: PrimitiveType.Types.F32,
        [WasmNumber.Type.F64]: PrimitiveType.Types.F64,
        // TODO add these
        [WasmNumber.Type.U32]: PrimitiveType.Types.I32,
        [WasmNumber.Type.U64]: PrimitiveType.Types.I64,
    };
    /**
     * @param name - formal name for type in target spec
     * @param size - memory size as measured by multiples of 32 bits
     */
    constructor(name, _size) {
        super(undefined);
        this.name = name;
        this._size = _size;
    }
    size() {
        return this._size;
    }
    /**
     * @override
     */
    getWasmTypeName() {
        return this.name;
    }
    /**
     * @override
     */
    check(type) {
        if (type == null)
            return false;
        if (type instanceof AnyType)
            return true;
        // Don't care about classes
        if (type instanceof ClassType) {
            type = type.getBaseType();
            if (!type)
                return false;
        }
        // All instances of PrimitiveType will be in the static Types map
        return this == type;
    }
    /**
     * @override
     */
    flatPrimitiveList() {
        return [this];
    }
    /**
     * @override
     */
    isUnit() {
        return false;
    }
    getBaseType() {
        return this;
    }
    toString() {
        return this.name.toUpperCase();
    }
}
/**
 * Datatype to describe function/macros
 */
export class ArrowType extends DataType {
    inputTypes;
    outputTypes;
    constructor(token, inputTypes, outputTypes) {
        super(token);
        this.inputTypes = inputTypes;
        this.outputTypes = outputTypes;
    }
    /**
     * @override
     */
    getWasmTypeName(name) {
        // Make sure it's valid compile-time type
        const badInp = this.inputTypes.find(t => !(t instanceof DataType));
        if (badInp)
            throw new error.SyntaxError('One of the input types is compile-time only', [badInp.token, this.token]);
        if (!this.outputTypes)
            throw new error.SyntaxError('Partial Arrow type used in wrong context', [this.token]);
        const badOut = this.outputTypes.find(t => !(t instanceof DataType));
        if (badOut)
            throw new error.SyntaxError('One of the input types is compile-time only', [badOut.token, this.token]);
        return `(func ${name} (param ${this.inputTypes.map(t => t.getWasmTypeName()).join(' ')}) (result ${this.outputTypes.map(t => t.getWasmTypeName()).join(' ')}))`;
    }
    /**
     * @override
     */
    check(type) {
        if (type instanceof AnyType)
            return true;
        if (!(type instanceof ArrowType))
            return false;
        return this.inputTypes.every((t, i) => t.check(type.inputTypes[i]))
            && (!this.outputTypes || this.outputTypes.every((t, i) => t.check(type.outputTypes[i])));
    }
    /**
     * Can this arrow be called with the given stack?
     * @param stack stack to check from
     * @returns true if types align
     */
    checkInputs(stack) {
        if (this.inputTypes.length === 0)
            return true;
        return stack.length >= this.inputTypes.length
            && stack
                .slice(-this.inputTypes.length)
                .every((v, i) => v.datatype && this.inputTypes[i].check(v.datatype));
    }
    checkInputTypes(types) {
        if (this.inputTypes.length === 0)
            return true;
        return types.length >= this.inputTypes.length
            && types
                .slice(-this.inputTypes.length)
                .every((t, i) => this.inputTypes[i].check(t));
    }
    /**
     * @override
     */
    flatPrimitiveList() {
        throw new Error('Invalid call: ArrowType.flatPrimitiveList()');
    }
    isUnit() {
        return false;
    }
    toString() {
        const inpTypes = this.inputTypes.map(t => t.toString()).join(' ');
        const outTypes = this.outputTypes.map(t => t.toString()).join(' ');
        return `( ${inpTypes} ) ( ${outTypes} ) Arrow`;
    }
}
/**
 * Reference for a value of a given type T
 *
 * Pointer stored on gc value stack, points to value in heap
 *
 * see: planning/brainstorm/ref_stack_vars.md
 */
export class RefType extends DataType {
    type;
    offsetBytes;
    /**
     * @param token location in source code
     * @param type type of value being referenced
     * @param offsetBytes When accessing a member of an object need to use reference
     */
    constructor(token, type, offsetBytes = 0) {
        super(token);
        this.type = type;
        this.offsetBytes = offsetBytes;
        // if (!type)
        //     console.log(this, new Error('wwww').stack);
    }
    /**
     * @override
     */
    check(type) {
        if (type instanceof AnyType)
            return true;
        // Interchangeable with base type
        return this.type.check(type);
    }
    /**
     * @override
     */
    getWasmTypeName() {
        return 'i32';
    }
    /**
     * @override
     */
    flatPrimitiveList() {
        // TODO this is bad, should just be [I32]
        // This should just give i32. For caller to get this functionality they should have to call this.type.flatPrimitiveList()
        return [this];
    }
    unpackRefs() {
        let offset = 0;
        return this.type.flatPrimitiveList().map(t => {
            const oldOffset = offset;
            if (t instanceof PrimitiveType)
                switch (t.name) {
                    case 'i32':
                    case 'f32':
                        offset += 4;
                        break;
                    case 'i64':
                    case 'f64':
                        offset += 8;
                        break;
                    default: throw new Error('wtf?');
                }
            else if (t instanceof EnumBaseType)
                offset += 8;
            else {
                // Otherwise it's a reference
                offset += 4;
                // console.log('RefType.fpl(): other type: ', t);
            }
            return new RefType(this.token, t, oldOffset);
        });
    }
    /**
     * Determines if objects of this type need to be stored on separate stack or not
     * @returns true if the type contains reference fields
     */
    hasRefs() {
        const bt = this.type instanceof ClassType ? this.type.getBaseType() : this.type;
        if (bt instanceof UnionType)
            return bt.types.some(t => t instanceof RefType || t instanceof RefRefType);
        // TODO arrow types, scary
        return false;
    }
    isUnit() {
        // It makes no sense for us to have a reference to a unit value
        return false; // this.type.isUnit();
    }
    size() {
        return 1;
    }
    toString() {
        return (this.type ? this.type.toString() : '_') + " Ref";
    }
}
/**
 * Reference to a pointer which is stored on rv_stack to an object managed by gc
 *
 * Stored in wasm locals, point to location on, gc rv stack, where lie pointers to objects on the heap
 *
 * see: planning/brainstorm/ref_stack_vars.md
 *
 * @depricated - I feel like conversion should be implicit
 */
export class RefRefType extends DataType {
    type;
    offsetBytes;
    /**
     * @param token location in source code
     * @param type type of value being referenced
     * @param offsetBytes When accessing a member of an object need to use reference
     */
    constructor(token, type, offsetBytes = 0) {
        super(token);
        this.type = type;
        this.offsetBytes = offsetBytes;
    }
    /**
     * @override
     */
    check(type) {
        if (type instanceof AnyType)
            return true;
        // Interchangeable with base type
        return this.type.check(type);
    }
    /**
     * @override
     */
    getWasmTypeName() {
        return 'i32';
    }
    flatPrimitiveList() {
        // TODO this is bad, should just be [I32]
        // let offset = 0;
        return [new RefType(this.token, this.type)];
        // return this.type.flatPrimitiveList().map(t => {
        //     const oldOffset = offset;
        //     if (t instanceof PrimitiveType)
        //         switch (t.name) {
        //             case 'i32': case 'f32': offset += 4; break;
        //             case 'i64': case 'f64': offset += 8; break;
        //             default: throw new Error('wtf?');
        //         }
        //     else
        //         // Otherwise it's a pointer
        //         offset += 4;
        //     return new RefType(this.token, t, oldOffset);
        // });
    }
    isUnit() {
        return false;
    }
    size() {
        return 1;
    }
    toString() {
        return this.type.toString() + " Ptr";
    }
}
/**
 * Matches anything within same enum type
 */
export class EnumBaseType extends DataType {
    subtypes;
    recursive;
    name;
    ns;
    /**
     * @param subtypes Enum members defined by this class
     */
    constructor(token, subtypes, recursive = false) {
        super(token);
        this.subtypes = subtypes;
        this.recursive = recursive;
        Object.entries(this.subtypes).forEach(([sym, t], i) => {
            t.parent = this;
            t.index = i;
            t.name = sym;
        });
    }
    /**
     * Update the subtype mappings
     */
    setSubtypes(subtypes) {
        this.subtypes = subtypes;
        Object.entries(subtypes).forEach(([sym, t], i) => {
            t.parent = this;
            t.index = i;
            t.name = sym;
        });
    }
    /**
     * Handle recursive types
     * When two type is recursive we only care about the non-recursive members
     * being compatible as that would make the recursive references compatible
     */
    _recCheck;
    /**
     * @override
     */
    check(type) {
        // Drop classes
        if (type instanceof ClassType)
            type = type.getBaseType();
        if (!type)
            return false;
        // Any type
        if (type instanceof AnyType)
            return true;
        // Yikes
        const isChild = () => Object.values(this.subtypes).some(t => t.check(type));
        const isCompat = () => {
            // Handle simple, non-recursive cases first
            if (!(type instanceof EnumBaseType && this.token === type.token))
                return false;
            // Recursive case
            if (this.recursive) {
                if (!this._recCheck)
                    this._recCheck = new WeakSet();
                if (this._recCheck.has(type))
                    return true;
                this._recCheck.add(type);
                const ret = Object.entries(this.subtypes).every(([name, st]) => st.type.check(type.subtypes[name].type));
                this._recCheck.delete(type);
                return ret;
            }
            // Non recursive
            return Object.entries(this.subtypes).every(([name, st]) => st.type.check(type.subtypes[name].type));
        };
        // console.log([this.name, (type as any).name], [this === type, isChild(), isCompat()]);
        return type && (this === type || isChild() || isCompat());
    }
    isUnit() {
        // return Object.values(this.subtypes).every(t => t.isUnit());
        return false;
    }
    flatPrimitiveList() {
        return [PrimitiveType.Types.I32, new RefType(this.token, null)];
    }
    getWasmTypeName() {
        // type index + ref address
        return 'i32';
    }
    getBaseType() {
        return this;
    }
    toString() {
        return `(: ${this.isRecursive()
            ? Object.keys(this.subtypes).join(' ') + ' '
            : `\n${Object.entries(this.subtypes).map(([sym, t]) => `${t.toString()} $${sym} =`).join('\n')}\n`}) enum`;
    }
    sortedSubtypes() {
        return Object.values(this.subtypes).sort((a, b) => a.index - b.index);
    }
    isRecursive() {
        return this.recursive;
    }
    getId(id) {
        return this.ns.getId(id);
    }
}
/**
 * Matches class defined within an enum
 */
export class EnumClassType extends DataType {
    type;
    name;
    recursive;
    /**
     * Metadata associating this class with the parent
     */
    parent;
    index;
    /**
     * @param token - Code where
     * @param type - Underlying data type stored
     * @param name - Name assigned to type
     * @param recursive - Does this class need to be stored on the heap?
     */
    constructor(token, type, name, recursive = false) {
        super(token);
        this.type = type;
        this.name = name;
        this.recursive = recursive;
    }
    /**
     * @override
     */
    check(type) {
        // Wildcard type
        if (type instanceof AnyType)
            return true;
        // Drop classes
        if (type instanceof ClassType)
            type = type.getBaseType();
        if (!type)
            return false;
        if (type === this.parent)
            return true;
        // if (type instanceof EnumClassType)
        //     console.log('st: ', [this.type, type.type],
        //         this.token === type.token,
        //         this.index === type.index,
        //         this.type.check(type.type));
        // else
        //     console.log(type.constructor);
        return type instanceof EnumClassType
            && this.token === type.token // corrolaries: same parent, same class id
            && this.index === type.index
            // TODO this is wrong
            && (this.type.check(type.type));
    }
    isUnit() {
        return false;
    }
    flatPrimitiveList() {
        // Is this right?
        return [PrimitiveType.Types.I32, new RefType(this.token, this.type)];
    }
    getWasmTypeName() {
        // Is this right?
        // type index + ref address
        return 'i32';
    }
    toString() {
        return `${this.type.toString()} ${this.name}#${this.index}_enum`;
    }
    getBaseType() {
        let ret = this.type;
        while (ret instanceof EnumClassType || ret instanceof ClassType)
            ret = ret.getBaseType();
        return ret;
    }
}
//# sourceMappingURL=datatypes.js.map