import * as value from '../value.js';
/**
 * This stores expressions that we can reason about
 * but can't completly eliminate from the code.
 *
 * For example, operations on user input and not constant-values
 *
 * @abstract
 * @class
 */
export class Expr extends value.Value {
    // State variable to prevent duplicated compilation
    _isCompiled = false;
    /**
     * @constructor
     * @param token - Source location
     */
    constructor(token) {
        super(token, value.ValueType.Expr, undefined);
    }
    /**
     * Would it be better to store the value in a local or inline it multiple times?
     * @returns true if performance would improve with caching false if inlining better
     * @virtual
     */
    get expensive() {
        return true;
    }
    /**
     * Exhaustive version of .children()
     * @returns all child nodes which don't have children
     */
    getLeaves() {
        let ret = new Set(this.children());
        let retLen = ret.size;
        do {
            retLen = ret.size;
            // console.log('v', retLen, [...ret][2]);
            ret = [...ret]
                .map(e => {
                const ret = e.children();
                return (!ret || ret.length === 0) ? e : ret;
            }).reduce((a, v) => {
                if (v instanceof Array) {
                    v.forEach(e => a.add(e));
                }
                else {
                    a.add(v);
                }
                return a;
            }, new Set());
        } while (retLen != ret.size);
        return [...ret];
    }
}
/**
 * Data Expressions
 * @abstract
 * @class
 */
export class DataExpr extends Expr {
    _datatype;
    /**
     * @param token - location in code
     * @param _datatype - Datatype for value
     */
    constructor(token, _datatype) {
        super(token);
        this._datatype = _datatype;
    }
    /**
     * @override
     */
    get expensive() {
        return false;
    }
    /**
     * @override
     */
    get datatype() {
        return this._datatype;
    }
    /**
     * @override
     */
    set datatype(t) {
        this._datatype = t;
    }
}
//# sourceMappingURL=expr.js.map