import { formatErrorPos } from "../tools/file_tools.js";
import WasmNumber from "./numbers.js";
// This handles Context-free grammar
// TODO optimize, use regex, char literals, etc.
// TODO this file is ugly
/**
 * Lexical type for token
 */
export var TokenType;
(function (TokenType) {
    TokenType[TokenType["String"] = 0] = "String";
    TokenType[TokenType["Number"] = 1] = "Number";
    TokenType[TokenType["ContainerOpen"] = 2] = "ContainerOpen";
    TokenType[TokenType["ContainerClose"] = 3] = "ContainerClose";
    TokenType[TokenType["Identifier"] = 4] = "Identifier";
    TokenType[TokenType["Block"] = 5] = "Block";
    TokenType[TokenType["Tuple"] = 6] = "Tuple";
})(TokenType || (TokenType = {}));
// Internal
export var ContainerType;
(function (ContainerType) {
    ContainerType[ContainerType["Curly"] = 0] = "Curly";
    ContainerType[ContainerType["Bracket"] = 1] = "Bracket";
    ContainerType[ContainerType["Paren"] = 2] = "Paren";
})(ContainerType || (ContainerType = {}));
// Generic token
export class LexerToken {
    token;
    type;
    position;
    file;
    static Type = TokenType;
    constructor(token, type, position, file) {
        this.token = token;
        this.type = type;
        this.position = position;
        this.file = file;
    }
    // TODO toString()
    locationString() {
        return formatErrorPos([{
                name: this.constructor.name + ' token',
                message: this.constructor.name + ' token',
                tokens: [this],
            }]);
    }
}
// Number literal
export class NumberToken extends LexerToken {
    value;
    constructor(token, position, file) {
        super(token, TokenType.Number, position, file);
        this.value = new WasmNumber().fromString(token);
    }
}
/**
 * The Token is initialized as a container open/close token
 * later it's converted to a block and the body is assigned
 */
export class BlockToken extends LexerToken {
    subtype;
    body;
    constructor(token, position, file) {
        const type = '[{('.includes(token) ? TokenType.ContainerOpen : TokenType.ContainerClose;
        super(token, type, position, file);
        this.subtype = [
            ContainerType.Curly, ContainerType.Curly,
            ContainerType.Bracket, ContainerType.Bracket,
            ContainerType.Paren, ContainerType.Paren,
        ]['{}[]()'.indexOf(token)];
    }
    isMacro() {
        return this.subtype === ContainerType.Paren
            && this.body
            && this.body.slice(0, 4).some(t => t.token === ':' || t.token === 'rec:');
    }
    toMacro() {
        // Not a macro
        if (!this.isMacro()) {
            this.type = TokenType.Tuple;
            return this;
        }
        const types = [];
        let recursive = false;
        let recursiveId = null;
        let i;
        const l = Math.min(4, this.body.length);
        for (i = 0; i < l; i++) {
            const t = this.body[i];
            if (t.token === ':')
                break;
            if (t.token === 'rec:' || t.token === 'fun:') {
                recursive = true;
                break;
            }
            if (t.token === 'rec' || t.token === 'fun')
                recursive = true;
            else if (t instanceof BlockToken)
                types.push(t);
            else if (t instanceof IdToken && t.isEscaped)
                if (recursiveId) {
                    throwParseError('Cannot have multiple fixed point identifiers', [this, recursiveId, t], this.file);
                }
                else {
                    recursive = true;
                    recursiveId = t;
                }
            else
                throwParseError('Invalid macro prefix', [this, t], this.file);
        }
        return new MacroToken(this.token, this.position, this.file, this.body.slice(i + 1), types, recursive, recursiveId);
    }
    /**
     * Get a list of identifiers referenced within the body of this block
     * @returns array of identifier tokens in the body
     */
    referencedIds() {
        const ret = [];
        this.body.forEach(tok => {
            if (tok instanceof BlockToken)
                ret.push(...tok.referencedIds());
            else if (tok instanceof IdToken)
                ret.push(tok);
        });
        return ret;
    }
}
export class MacroToken extends LexerToken {
    body;
    types;
    recursive;
    recursiveId;
    constructor(token, position, file, body, types, recursive = false, recursiveId = null) {
        super(token, TokenType.Block, position, file);
        this.body = body;
        this.types = types;
        this.recursive = recursive;
        this.recursiveId = recursiveId;
    }
}
export class IdToken extends LexerToken {
    value;
    isEscaped;
    constructor(token, position, file) {
        super(token, TokenType.Identifier, position, file);
        this.isEscaped = this.token[0] === '$';
        const unescaped = this.isEscaped ? this.token.slice(1) : this.token;
        // TODO handle when token is "." or "..." or "..abc" this could be useful?
        this.value = unescaped.split('.');
    }
}
/**
 * Describes tokenized string
 *
 * @param token - lexical token string
 * @param options - override/extend defaults
 */
function toToken(token, position, file) {
    // Trim whitespace
    token = token.trim();
    // Filter empty tokens
    if (token.length === 0)
        return null;
    // String
    if (token[0] === '"')
        return new LexerToken(token.substring(1, token.length - 1), TokenType.String, position, file);
    // Number (note this makes NaN an identifier)
    if (!isNaN(parseFloat(token)))
        return new NumberToken(token, position, file);
    // Separators
    if ('{}[]()'.includes(token))
        return new BlockToken(token, position, file);
    // Identifier
    return new IdToken(token, position, file);
}
/**
 * Generates a list of tokens from given program source
 *
 * @param src - program source code
 * @param file - file name/path
 * @returns - List of tokens
 */
export function lex(src, file) {
    let i = 0, prev = 0;
    const ret = [];
    // Add token to return
    const addToken = (s) => ret.push(toToken(s, i, file));
    // Find end of string
    const endStr = () => {
        // Determine if quote is escaped
        function isEscaped(s, i) {
            let e = false;
            while (s[--i] === '\\')
                e = !e;
            return e;
        }
        // Find end of string
        while (++i < src.length)
            if (src[i] === '"' && !isEscaped(src, i))
                break;
    };
    // For each char...
    while (i < src.length) {
        // Separator
        if ('[]{}()'.includes(src[i])) {
            addToken(src.substring(prev, i));
            addToken(src[i]);
            // Line-Comment
        }
        else if (src[i] === '#') {
            addToken(src.substring(prev, i));
            while (i < src.length && src[i] !== '\n')
                i++;
            // End of token
        }
        else if ([' ', '\t', '\n'].includes(src[i])) {
            addToken(src.substring(prev, i));
            // String literal
        }
        else if (['"', "'"].includes(src[i])) {
            addToken(src.substring(prev, i));
            prev = i;
            endStr();
            i++;
            addToken(src.substring(prev, i));
            // Middle of a token
        }
        else {
            i++;
            continue;
        }
        // Next character in token
        prev = ++i;
    }
    // EOF is a separator
    if (i !== prev)
        addToken(src.substring(prev, i));
    // Return list of token objects
    return ret.filter(Boolean);
}
/**
 * Throw syntax error
 *
 * @param message - reason for error
 * @param tokens - problematic tokens
 * @param file - file name/path
 */
function throwParseError(message, tokens, file) {
    // TODO convert to constructor
    throw {
        type: 'SyntaxError',
        name: 'LexerError',
        message,
        tokens,
        stack: new Error(),
        file,
    };
}
/**
 * Generates a parse tree from list of tokens
 * Really only benefit here is that this collapses containers
 *
 * @param code - code to scan
 * @param file - file name/path
 */
export default function scan(code, file) {
    // Parse tokens
    const tokens = lex(code, file);
    const ret = [];
    tokens.forEach(tok => {
        // Collapse containers
        if (tok.type == TokenType.ContainerClose) {
            // Index of most recent container
            // TODO use .reverse+findIndex?
            let ind;
            for (ind = ret.length - 1; ind >= 0; ind--)
                if (ret[ind].type === TokenType.ContainerOpen)
                    break;
            // No openers
            if (ind === -1)
                throwParseError('Unexpected symbol ' + tok.token, [tok]);
            // Not matching
            const parent = ret[ind];
            if (parent.subtype !== tok.subtype)
                throwParseError(`Container mismatch ${parent.token} vs. ${tok.token}`, [parent, tok]);
            // Collapse body
            parent.body = ret.splice(ind + 1);
            ret[ind] = parent.toMacro();
        }
        else {
            ret.push(tok);
        }
    });
    return ret;
}
/**
 * Used for code originating within the compiler
 */
export const internalToken = new IdToken('compiler internal', 0, '/dev/null');
//# sourceMappingURL=scan.js.map