import { LexerToken } from './scan.js';
import * as types from './datatypes.js'; // If we actually have to import datatypes here it will not work
// TODO should move these to /expr/value.ts so that cyclic imports are less anal
/**
 * In this context, Values are like nodes on an AST, but also used to simplify constexprs/partial evaluation
 */
/**
 * Enum to represent what syntactic type is currently on the stack
 */
// TODO should probably use Classes/Inheritance instead of this
export var ValueType;
(function (ValueType) {
    ValueType[ValueType["Macro"] = 0] = "Macro";
    ValueType[ValueType["Data"] = 1] = "Data";
    ValueType[ValueType["Type"] = 2] = "Type";
    ValueType[ValueType["Id"] = 3] = "Id";
    ValueType[ValueType["Expr"] = 4] = "Expr";
    ValueType[ValueType["Fxn"] = 5] = "Fxn";
    ValueType[ValueType["Str"] = 6] = "Str";
    ValueType[ValueType["Ns"] = 7] = "Ns";
})(ValueType || (ValueType = {}));
// TODO should be abstract
/**
 * Generic value base class
 */
export class Value {
    _datatype;
    /// Source location in code
    token;
    /// Node type
    type;
    /// Relevant Value
    value;
    // Enum relevant here but ts doesn't let me use it
    static Type = ValueType;
    constructor(token, type, value, _datatype) {
        this._datatype = _datatype;
        this.token = token;
        this.type = type;
        this.value = value;
    }
    /**
     * Do we know it at compile time?
     */
    isConstExpr() {
        return this.type !== ValueType.Expr;
    }
    // out(ctx: ModuleManager, fun?: expr.FunExpr): string {
    //     return expr.fromDataValue([this]).map(e => e.out(ctx, fun)).join(' ');
    // }
    /**
     * Name for type of this value
     */
    typename() {
        return ValueType[this.type];
    }
    get datatype() {
        return this._datatype || types.SyntaxType.ValueTypes[this.type];
    }
    /**
     * @depricated
     */
    set datatype(t) {
        this._datatype = t;
    }
    get expensive() {
        return false;
    }
    children() {
        return [];
    }
}
/**
 * Data with a user-level type, this includes numbers, tuples and classes
 */
export class DataValue extends Value {
    type = ValueType.Data;
    constructor(token, type, value) {
        super(token, ValueType.Data, value, type);
    }
    get datatype() {
        return this._datatype;
    }
    set datatype(t) {
        this._datatype = t;
    }
}
/**
 * Primitive data, native wasm types
 */
export class NumberValue extends DataValue {
    constructor(token, wasmNumber) {
        super(token, types.PrimitiveType.typeMap[wasmNumber.type], wasmNumber);
    }
    /**
     * See code in expr/expr.ts
     */
    out() {
        return this.value.toWAST();
    }
    /**
     * See code in expr/expr.ts
     */
    children() {
        return [];
    }
    get datatype() {
        return this._datatype;
    }
    set datatype(t) {
        this._datatype = t;
    }
}
/**
 * Escaped Identifier
 */
export class IdValue extends Value {
    isGlobal;
    type = ValueType.Id;
    constructor(token, isGlobal = false) {
        super(token, ValueType.Id, token.value);
        this.isGlobal = isGlobal;
    }
    /**
     * Convert Id's to values
     */
    deref(ctx) {
        return ctx.getId(this.token.value);
    }
}
/**
 * Packed values
 */
export class TupleValue extends DataValue {
    constructor(token, values, datatype) {
        const type = datatype || new types.TupleType(token, values.map(v => v.datatype || null));
        super(token, type, values);
    }
    children() {
        return [].concat(...this.value.map(v => v.children && v.children()));
    }
    out(ctx, fun) {
        return this.value.map(v => v.out(ctx, fun)).join('');
    }
    get datatype() {
        return this._datatype;
    }
    set datatype(t) {
        this._datatype = t;
    }
}
// Note that there is no need for a UnionValue class because there are no instances of unions
// unions will only be needed as types for opaque?-expressions
/**
 * String literal, not data
 */
export class StrValue extends Value {
    constructor(token, str) {
        if (str)
            super(token, ValueType.Str, typeof str === 'string' ? str : new TextDecoder().decode(str));
        else if (token.type === LexerToken.Type.String)
            super(token, ValueType.Str, token.token);
        else
            throw new Error('Invalid call');
    }
}
/**
 * Set of identifiers
 */
export class NamespaceValue extends Value {
    type = ValueType.Ns;
    constructor(token, ns) {
        super(token, ValueType.Ns, ns);
    }
}
//# sourceMappingURL=value.js.map